// Karplus-Strong string synthesis algorithm implementation, spits out the file to stdout as a WAV file
// if you use mpv, you can redirect it to mpv directly: `./karplus_strong <freq> | mpv -`
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <bsd/stdlib.h> // arc4random (link with -lbsd)

struct wav_header // not the most portable way really
{
	char chunk_id[4];
	uint32_t chunk_size;
	char format[4];
	char subchunk_id[4];
	uint32_t subchunk1_size;
	uint16_t audio_format;
	uint16_t num_channels;
	uint32_t sample_rate;
	uint32_t byte_rate;
	uint16_t block_align;
	uint16_t bits_per_sample;
	char subchunk2_id[4];
	uint32_t subchunk2_size;
};

void wav_populate_header(struct wav_header* wavh, size_t total_samples, size_t channels, size_t bits_per_sample, size_t sample_rate)
{
	memcpy(wavh->chunk_id, "RIFF", 4);
	memcpy(wavh->format, "WAVE", 4);
	memcpy(wavh->subchunk_id, "fmt ", 4);

	wavh->chunk_size = (sizeof(struct wav_header) + total_samples) - 8;
	wavh->subchunk1_size = 16;
	wavh->audio_format = 1;
	wavh->num_channels = 1;
	wavh->sample_rate = sample_rate;
	wavh->byte_rate = (sample_rate * channels * bits_per_sample) / 8;
	wavh->block_align = (channels * bits_per_sample) / 8;
	wavh->bits_per_sample = bits_per_sample;

	memcpy(wavh->subchunk2_id, "data", 4);
	wavh->subchunk2_size = (total_samples * channels * bits_per_sample) / 8;
}

void wav_create_file(struct wav_header* wavh, FILE* wav_fp, int16_t* data, size_t total_samples)
{
	fwrite(wavh, sizeof(struct wav_header), 1, wav_fp); // write WAV header to file
	fwrite(data, sizeof(int16_t), total_samples, wav_fp); // ...and finally the audio data
}

int main(int argc, char** argv)
{
	if (argc < 2) {
		fprintf(stderr, "frequency argument expected\n");
		return 1;
	}

	const size_t SAMPLE_RATE = 44100;
	const size_t freq = strtoull(argv[1], NULL, 0); // you should check for errors
	const size_t seconds = 3;
	const size_t total_samples = seconds * SAMPLE_RATE;
	const size_t period = SAMPLE_RATE / freq;

	int16_t* noise_table = malloc(period * sizeof(int16_t)); // wavetable
	int16_t* buffer = malloc(total_samples * sizeof(int16_t)); // PCM buffer
	// fill wavetable with noise, it doesn't have to be this function, rand() over all elements would suffice
	arc4random_buf(noise_table, period);

	// algorithm
	size_t prev_val = 0;
	for (size_t n = 0; n < total_samples; ++n) {
		size_t cs = n % period;
		noise_table[cs] = (noise_table[cs] + prev_val) * 0.5;
		buffer[n] = noise_table[cs];
		prev_val = noise_table[cs];
	}

	// finally write the WAV file
	struct wav_header wavh = {};
	wav_populate_header(&wavh, total_samples, 1, 16, SAMPLE_RATE);
	wav_create_file(&wavh, stdout, buffer, total_samples);

	free(noise_table);
	free(buffer);

	return 0;
}
