;;;; Native words definitions
(in-package :pila)

(defmacro rpndefnat (name &body body)
  "Defines a new native word and adds it to the dictionary"
  `(setf (gethash ,name *rpn-dict*)
         (make-native-word ,name (lambda () ,@body))))

;;; Binary arithmetic words
(defmacro binop (operator-function &key operator-type)
  "Wrapper that pops two values from the stack and applies those values
  to the function specified by OP"
  `(let ((y (vector-pop *rpn-stack*))
         (x (vector-pop *rpn-stack*)))
      (check-type x ,operator-type)
      (check-type y ,operator-type)
      (vector-push-extend (funcall ,operator-function x y) *rpn-stack*)))

(rpndefnat "+"
  (binop #'+ :operator-type number))
(rpndefnat "*"
  (binop #'* :operator-type number))
(rpndefnat "-"
  (binop #'- :operator-type number))
(rpndefnat "/"
  (binop #'/ :operator-type number))
(rpndefnat "mod"
  (binop #'mod :operator-type number))

(rpndefnat "#t" (vector-push-extend *true-value* *rpn-stack*))
(rpndefnat "#f" (vector-push-extend *false-value* *rpn-stack*))

;;; Stack manipulation words
(rpndefnat "dup"
  (vector-push-extend (vector-last *rpn-stack*) *rpn-stack*))

(rpndefnat "drop"
  (vector-pop *rpn-stack*))

(rpndefnat "swap"
  (let ((val2 (vector-pop *rpn-stack*))
        (val1 (vector-pop *rpn-stack*)))
    (vector-push-extend val2 *rpn-stack*)
    (vector-push-extend val1 *rpn-stack*)))
    
(rpndefnat "s?"
  (let ((stack-len (length *rpn-stack*)))
    (cond
      ((zerop stack-len)
       (format t "stack is empty~%"))
      (t
       (format t "stack contents (~d):~%" stack-len)
       (loop for e across *rpn-stack* do
             (format t "~a " e)
             (finish-output))
       (terpri)))))

(rpndefnat "s!"
  (setf *rpn-stack* (make-dynarray +rpn-stack-size+)))

(rpndefnat "not"
  (let ((val (vector-pop *rpn-stack*)))
    (vector-push-extend (rpn-bool-not val) *rpn-stack*)))

(rpndefnat "and" (binop #'rpn-bool-and :operator-type boolean-word))
(rpndefnat "or" (binop #'rpn-bool-or :operator-type boolean-word))

(rpndefnat "="
  (let ((rhs (vector-pop *rpn-stack*))
        (lhs (vector-pop *rpn-stack*)))
    (vector-push-extend
      (native-bool-to-rpn (= lhs rhs)) *rpn-stack*)))

(rpndefnat "<"
  (let ((rhs (vector-pop *rpn-stack*))
        (lhs (vector-pop *rpn-stack*)))
    (vector-push-extend
      (native-bool-to-rpn (< lhs rhs)) *rpn-stack*)))

(rpndefnat "eval"
  (eval-word (vector-pop *rpn-stack*)))

; null word, used as a placeholder
(rpndefnat "$$" nil)

(rpndefnat ":"
  (setf *compile-mode* t))

(rpndefnat ";"
  (setf *current-def-word* nil
        *compile-mode* nil))

(rpndefnat "let"
  (setf (gethash nil *variables*) nil))

(rpndefnat "print-top"
  (format t "~a~%" (vector-last *rpn-stack*)))

(rpndefnat "read-number"
  (princ "> ")
  (finish-output)
  (vector-push-extend (parse-num (read-line *standard-input*)) *rpn-stack*))

(rpndefnat "if"
  (let ((false-branch (vector-pop *rpn-stack*))
        (true-branch (vector-pop *rpn-stack*))
        (ccond (vector-pop *rpn-stack*)))
    (eval-word (if (rpn-bool-true-p ccond)
                   true-branch
                   false-branch))))

(rpndefnat "dotimes"
  (let ((loop-body (vector-pop *rpn-stack*))
        (times (vector-pop *rpn-stack*)))
    (dotimes (n times)
      (eval-word loop-body))))

(rpndefnat "while"
  (let ((loop-body (vector-pop *rpn-stack*))
        (while-condition *true-value*))
    (loop while (rpn-bool-true-p while-condition) do
          (eval-word loop-body)
          (setf while-condition (vector-pop *rpn-stack*)))))

(rpndefnat "words"
  (maphash
    (lambda (word-name word)
      (let ((word-type (if (typep word 'native-word) "Nat" "User")))
        (format t "~8a~a~%" word-type word-name)))
    *rpn-dict*)
  (format t "~d words on the dictionary~%" (hash-table-count *rpn-dict*)))

(rpndefnat "exit"
  (let ((exit-code (vector-pop *rpn-stack*)))
    (uiop:quit exit-code)))
