package main

import (
	"bufio"
	"fmt"
	"golang.org/x/term"
	"os"
	"strings"
)

type Word struct {
	name string
	impl func(stack *Stack) error
}

var Dictionary = map[string]Word{
	"+":    Word{impl: natAdd},
	"*":    Word{impl: natMul},
	"-":    Word{impl: natSub},
	"/":    Word{impl: natDiv},
	"mod":  Word{impl: natMod},
	".":    Word{impl: natPrint},
	"s?":   Word{impl: natPrintStack},
	"dup":  Word{impl: natDup},
	"drop": Word{impl: natDrop},
	"rnd":  Word{impl: natRnd},
	"bye":  Word{impl: natBye},
}

func isNumberAscii(chr byte) bool {
	return chr >= '0' && chr <= '9'
}

func isIdent(chr byte) bool {
	return (chr >= 'A' && chr <= 'Z') ||
		(chr >= 'a' && chr <= 'z') ||
		(chr >= '!' && chr <= '/') ||
		(chr >= ':' && chr <= '`')
}

func parseNumber(s string, cur *int) StackType {
	var num int
	const digits = "0123456789abcdf"
	for _, c := range s {
		bc := byte(c)
		if !isNumberAscii(bc) {
			*cur--
			break
		}
		idx := strings.IndexByte(digits, bc)
		if idx < 0 { // we should't reach this place
			panic("unexpected non-digit character")
		}
		num = num*10 + idx
		*cur++
	}
	return StackType(num)
}

func parseWord(s string, cur *int) string {
	for i, c := range s {
		if !isIdent(byte(c)) {
			return s[:i]
		}
		*cur++
	}
	return s
}

func lookupWord(wordName string) (*Word, error) {
	word, ok := Dictionary[wordName]
	if !ok {
		return nil, fmt.Errorf("no such word: `%s`", wordName)
	}
	return &word, nil
}

func eval(stack *Stack, prog string) {
	for i := 0; i < len(prog); i++ {
		bc := byte(prog[i])
		if bc == ' ' || bc == '\t' || bc == '\n' || bc == '\r' {
			continue
		}
		if isNumberAscii(bc) {
			num := parseNumber(prog[i:], &i)
			stack.push(num)
		} else {
			wordImpl, err := lookupWord(parseWord(prog[i:], &i))
			if err != nil {
				fmt.Printf("Error: %s\n", err)
				break
			}
			if err := wordImpl.impl(stack); err != nil {
				fmt.Printf("Error: %s\n", err)
			}
		}
	}
}

func main() {
	stack := new(Stack)
	stack.data = make([]StackType, 0, StackSize)

	Dictionary["words"] = Word{impl: natHelp}
	Dictionary["s!"] = Word{
		impl: func(stack *Stack) error {
			stack.wipe()
			return nil
		},
	}

	is_tty := term.IsTerminal(int(os.Stdin.Fd()))
	scanner := bufio.NewScanner(os.Stdin)
	if is_tty {
		fmt.Println("RPN calculator - Type Ctrl-D or \"bye\" to exit")
	}

	for {
		if is_tty {
			fmt.Print("> ")
		}
		if !scanner.Scan() {
			if err := scanner.Err(); err != nil {
				panic(err)
			}
			break
		}
		eval(stack, scanner.Text())
	}
}
